local rt = RoosterTools
local util = RoosterToolsUtil

-------------------------------------------------------------------------------
-- RoosterToolsComm 
--   All addon communications.
-------------------------------------------------------------------------------
RoosterToolsComm = {}

RoosterToolsComm.TYPES = {
    -- Request from another guild user to receive data from this instance
    GUILD_SYNC_REQ = "RTSyncRequest",

    -- Received sync data from another guild member. 
    SYNC_GROUP_RESP = "RTGroupResp",
    SYNC_CHAR_RESP = "RTCharResp",
    SYNC_MAIN_RESP = "RTMainResp",
    SYNC_ALIAS_RESP = "RTAliasResp"
}

RoosterToolsComm.METHOD = {
    WHISPER = "WHISPER",
    GUILD = "GUILD"
}

local function newerThan(tbl, t)
    local g = {}

    for k, v in pairs(tbl) do
        if ((v.updated and v.updated > t) or (v.time and v.time > t)) then
            g[k] = v
        end
    end

    return g
end

function RoosterToolsComm:Send(type, data, method, target)
    local serialized = rt:Serialize({
        ["version"] = 1,
        ["data"] = data
    })
    rt:SendCommMessage(type, serialized, method, target, "BULK", function(a, b, c)
        RoosterToolsUtil.logDebug(target or method .. " " .. type .. " " .. b .. " " .. c)
    end)
end

function RoosterToolsComm.GuildSyncReqHandler(prefix, data, dist, target, ...)
    local recv = time()

    -- Skip handling request from the current player.
    if (target == rt.player.playerName) then
        return
    end

    local success, d = rt:Deserialize(data)
    if (not success) then
        return
    end

    local sync = rt.db.sync

    local syncIt = function(type, data)
        local toonName = target

        -- Override the toon's name with their main so we can avoid syncing data that isn't necessary. 
        local main = rt.db.mains[target]
        if (main) then
            toonName = main.name
        end

        for k, v in pairs(data) do
            local key = toonName .. k
            local lastResponse = 0
            if (sync[key]) then
                lastResponse = sync[key].lastResponse
                sync[key].lastResponse = recv
            else
                sync[key] = {
                    lastResponse = recv
                }
            end

            if ((v.updated and v.updated > lastResponse) or (v.time and v.time > lastResponse)) then
                local t = {}
                t[k] = v

                RoosterToolsComm:Send(type, t, RoosterToolsComm.METHOD.WHISPER, target)
            end
        end
    end

    syncIt(RoosterToolsComm.TYPES.SYNC_CHAR_RESP, rt.db.characters)
    syncIt(RoosterToolsComm.TYPES.SYNC_GROUP_RESP, rt.db.groups)
    syncIt(RoosterToolsComm.TYPES.SYNC_MAIN_RESP, rt.db.mains)
    syncIt(RoosterToolsComm.TYPES.SYNC_ALIAS_RESP, rt.db.aliases)
end

function RoosterToolsComm.SyncCharRespHandler(prefix, data, dist, target, ...)
    local success, d = rt:Deserialize(data)
    if (not success) then
        return
    end

    if (type(d) ~= "table") then
        return
    end

    if (d.version == 1) then
        if (d.data) then
            for k, v in pairs(d.data) do
                if (not rt.db.characters[k]) then
                    util.logInfo("Add Character " .. k .. " from " .. target) 
                    rt.db.characters[k] = v
                elseif (rt.db.characters[k].updated < v.updated) then
                    util.logInfo("Update Character " .. k .. " from " .. target) 
                    rt.db.characters[k] = v
                end
            end
        end
    end
end

local function mergeGroup(old, new) 
    old.created = math.min(old.created, new.created)

    old.updated = new.updated
    old.type = new.type

    if (old.guildGroupTime and new.guildGroupTime) then
        old.guildGroupTime = math.min(old.guildGroupTime, new.guildGroupTime)
    end

    for k, newInstance in pairs(new.instances) do
        -- There is a window where we could dupe an instance due to the inability to uniquely id it
        -- To avoid a collision, we'll just avoid taking extremely new data in the case of a sync.
        if (newInstance.created < (time() + 60)) then
            -- find the instance in old
            local foundInstance = nil
            for k, oldInstance in pairs(old.instances) do
                if (oldInstance.instanceName == newInstance.instanceName 
                    and oldInstance.difficultyName == newInstance.difficultyName 
                    and oldInstance.updated > newInstance.created) then
                    oldInstance.created = math.min(oldInstance.created, newInstance.created)
                    oldInstance.updated = newInstance.updated
                    foundInstance = oldInstance
                end
            end

            if (not foundInstance) then
                table.insert(old.instances, newInstance)
            end
        end
    end

    for k, newEncounter in pairs(new.encounters) do
        -- There is a window where we could dupe an instance due to the inability to uniquely id it
        -- To avoid a collision, we'll just avoid taking extremely new data in the case of a sync.
        if (newEncounter.ended and newEncounter.started < (time() + 60)) then
            -- find the instance in old
            local foundEncounter = nil
            for k, oldEncounter in pairs(old.encounters) do
                if (oldEncounter.encounterName == newEncounter.encounterName 
                    and oldEncounter.difficultyID == newEncounter.difficultyID 
                    and oldEncounter.instance == newEncounter.instance 
                    and ((oldEncounter.started >= newEncounter.started and oldEncounter.started <= newEncounter.ended) or 
                         (newEncounter.started >= oldEncounter.started and newEncounter.started <= oldEncounter.ended))) then
                            oldEncounter.started = math.min(oldEncounter.started, newEncounter.started)
                            oldEncounter.ended = math.max(oldEncounter.ended, newEncounter.ended)
                    foundEncounter = oldEncounter
                end
            end

            if (not foundEncounter) then
                table.insert(old.encounters, newEncounter)
            end
        end
    end

    for k, newCharacter in pairs(new.characters) do
        local oldCharacter = old.characters[k]
        if (oldCharacter) then
            if (newCharacter.guildName) then
                oldCharacter.guildName = newCharacter.guildName
            end
            
            for k, newState in pairs(newCharacter.states) do
                local foundState = nil
                for k, oldState in pairs(oldCharacter.states) do
                    if (oldState.state == newState.state and oldState.time == newState.time) then
                        foundState = oldState.state
                        break
                    end
                end

                if (not foundState) then
                    table.insert(oldCharacter.states, newState)
                end
            end
        else
            old.characters[k] = newCharacter
        end
    end
end

function RoosterToolsComm.SyncGroupRespHandler(prefix, data, dist, target, ...)
    local success, d = rt:Deserialize(data)
    if (not success) then
        return
    end

    if (type(d) ~= "table") then
        return
    end

    if (d.version == 1) then
        if (d.data) then
            for k, v in pairs(d.data) do
                if (not rt.db.groups[k]) then
                    util.logInfo("Add Group " .. k .. " from " .. target) 
                    rt.db.groups[k] = v
                elseif (rt.db.groups[k].updated < v.updated and v.created > (time() + 60)) then
                    util.logInfo("Update Group " .. k .. " from " .. target) 
                    rt.db.groups[k] = mergeGroup(rt.db.groups[k], v)
                end
            end
        end
    end
end

function RoosterToolsComm.SyncMainRespHandler(prefix, data, dist, target, ...)
    local success, d = rt:Deserialize(data)
    if (not success) then
        return
    end

    if (type(d) ~= "table") then
        return
    end

    if (d.version == 1) then
        if (d.data) then
            for k, v in pairs(d.data) do
                if (not rt.db.mains[k]) then
                    util.logInfo("Add Main " .. k .. " from " .. target) 
                    rt.db.mains[k] = v
                elseif (rt.db.mains[k].time < v.time) then
                    util.logInfo("Update Main " .. k .. " from " .. target) 
                    rt.db.mains[k] = v
                end
            end
        end
    end
end

function RoosterToolsComm.SyncAliasRespHandler(prefix, data, dist, target, ...)
    local success, d = rt:Deserialize(data)
    if (not success) then
        return
    end

    if (type(d) ~= "table") then
        return
    end

    if (d.version == 1) then
        if (d.data) then
            for k, v in pairs(d.data) do
                if (not rt.db.aliases[k]) then
                    util.logInfo("Add Alias " .. k .. " from " .. target) 
                    rt.db.aliases[k] = v
                elseif (rt.db.aliases[k].updated < v.updated) then
                    util.logInfo("Update Alias " .. k .. " from " .. target) 
                    rt.db.aliases[k] = v
                end
            end
        end
    end
end

-- Register for comms
function RoosterToolsComm:OnInitialize()
    rt:RegisterComm(RoosterToolsComm.TYPES.GUILD_SYNC_REQ, RoosterToolsComm.GuildSyncReqHandler)
    rt:RegisterComm(RoosterToolsComm.TYPES.SYNC_CHAR_RESP, RoosterToolsComm.SyncCharRespHandler)
    rt:RegisterComm(RoosterToolsComm.TYPES.SYNC_GROUP_RESP, RoosterToolsComm.SyncGroupRespHandler)
    rt:RegisterComm(RoosterToolsComm.TYPES.SYNC_MAIN_RESP, RoosterToolsComm.SyncMainRespHandler)
    rt:RegisterComm(RoosterToolsComm.TYPES.SYNC_ALIAS_RESP, RoosterToolsComm.SyncAliasRespHandler)
end
